#ifndef _ALTERA_DMA_H
#define _ALTERA_DMA_H
#include "altera_dma_regs.h"
#include "altera_dma_cmd.h"
#include "pcie40_driver_common.h"

#define ALTERA_DMA_DRIVER_NAME    "Altera DMA"
#define ALTERA_DMA_DEVFILE        "altdma"

#define P40_MAX_BAR 5
#define DMA_BARS  ((1<<0)|(1<<4))  // for BAR0 and BAR4

#define CTL_STS_BITS                    0x0100

#define TIMEOUT_THRESH                  0xFFFF
#define MAX_NUM_DWORDS                  0x7FFFF//4096


struct dma_descriptor {
    u32 src_addr_ldw;
    u32 src_addr_udw;
    u32 dest_addr_ldw;
    u32 dest_addr_udw;
    u32 ctl_dma_len;
    u32 reserved[3];
} __attribute__ ((packed));




struct lite_dma_header {
    volatile u32 flags[128];
} __attribute__ ((packed));


struct lite_dma_desc_table {
    struct lite_dma_header header;
    struct dma_descriptor descriptors[ALTERA_DMA_DESCRIPTOR_NUM];
} __attribute__ ((packed));

struct pcie40_dma_state {
    struct pci_dev *pci_dev;

    u8 revision;
    u8 irq_pin;
    int irq_count;
    char msi_enabled;
    u8 irq_line;
    char dma_capable;

    void * __iomem bar[P40_MAX_BAR];
    size_t bar_length[P40_MAX_BAR];

    struct lite_dma_desc_table *lite_table_rd_cpu_virt_addr;
    struct lite_dma_desc_table *lite_table_wr_cpu_virt_addr;
 
    dma_addr_t lite_table_rd_bus_addr; 
    dma_addr_t table_rd_bus_addr; 
    dma_addr_t lite_table_wr_bus_addr;
    dma_addr_t table_wr_bus_addr;

    int numpages;
    u8 *rp_rd_buffer_virt_addr;
    dma_addr_t rp_rd_buffer_bus_addr;
    u8 *rp_wr_buffer_virt_addr;
    dma_addr_t rp_wr_buffer_bus_addr;

    dev_t cdevno;
    struct cdev cdev;

    int user_pid;
    struct task_struct *user_task;
    wait_queue_head_t wait_q;
    atomic_t status;

    struct dma_status dma_status;
    struct pcie40_state *common;
};

static int scan_bars(struct pcie40_dma_state *bk_ptr, struct pci_dev *dev) __init;
static int map_bars(struct pcie40_dma_state *bk_ptr, struct pci_dev *dev) __init;
static int dma_test(struct pcie40_dma_state *bk_ptr, struct pci_dev *dev);
//static irqreturn_t dma_isr(int irq, void *dev_id);

int altera_pci_probe(struct pci_dev *dev, const struct pci_device_id *id) ;

static int scan_bars(struct pcie40_dma_state *bk_ptr, struct pci_dev *dev);
void altera_pci_remove(struct pci_dev *dev) ;

static int init_chrdev (struct pcie40_dma_state *bk_ptr) __init;

ssize_t altera_dma_read(struct file *file, char __user *buf, size_t count, loff_t *pos);
ssize_t altera_dma_write(struct file *file, char __user *buf, size_t count, loff_t *pos);
ssize_t altera_dma_rw(struct file *file, char __user *buf, size_t count, loff_t *pos, int read);
int altera_dma_open(struct inode *inode, struct file *file);
int altera_dma_release(struct inode *inode, struct file *file);

static long altera_dma_ioctl (struct file *filp, unsigned int cmd, unsigned long arg);



#endif /* _ALTERA_DMA_H */
